/*
 * Scratchpad MMU - Test program module 0
 * Copyright (C) Jack Whitham 2009
 * http://www.jwhitham.org.uk/c/smmu.html
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation
 * (version 2.1 of the License only).
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */
#undef NDEBUG
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <assert.h>
#include "simif.h"
#include "print.h"

#ifdef __MICROBLAZE__
#include "mb_interface.h"
#include "xparameters.h"
#include "hwdriver.h"
#endif

#define TIMER_THRESHOLD         4

//#define VERBOSE 1
#define MAX_FREE_LIST_SIZE      100
#define NUM_CYCLES              1000000

unsigned rng ( void ) ;
void reset_rng ( void ) ;

data_t  SHADOW [ MM_SIZE ] ;
unsigned cycle, unique, open_ok, open_failed ;
unsigned active_hwm, close_ok ;


typedef struct Active_List_Entry_struct
{
    static_cache_ref    ref ;
    size_t              size , address , copy_location ;
} Active_List_Entry ;

typedef struct Free_List_Entry_struct
{
    size_t              size , copy_location ;
} Free_List_Entry ;

unsigned            active_list_size = 0 ;
unsigned            free_list_size = 0 ;
Free_List_Entry *   free_list;
Active_List_Entry * active_list;



void Cache_Test () ;

static void Check ( int call_exit )
{
    size_t i ;
    unsigned errors = 0 ;

    for ( i = 0 ; i < MM_SIZE ; i ++ )
    {
        data_t shadow = SHADOW [ i ] ;
        data_t sw_sc = TEST_SW_LOAD ( i ) ;
        data_t hw_sc = sw_sc ;
        int sw_fast = TEST_SW_Get_Result_Timer () < TIMER_THRESHOLD ;
        int hw_fast = sw_fast ;
        
#ifdef __MICROBLAZE__
        hw_sc = TEST_HW_LOAD ( i ) ;
        hw_fast = TEST_HW_Get_Result_Timer () < TIMER_THRESHOLD ;
#endif

        if (( sw_sc != shadow )
        || ( hw_sc != shadow )
        || ( hw_fast != sw_fast ))
        {
            printf ( "Inconsistency detected.\r\n"
                     "  address = 0x%08x" , (unsigned) i) ;
            printf ( "\r\n  should be = 0x%08x", shadow ) ;
            printf ( "\r\n  software says = 0x%08x", sw_sc ) ;
            if ( sw_fast ) 
                printf ( " (fast)" ) ;
#ifdef __MICROBLAZE__
            printf ( "\r\n  hardware says = 0x%08x", hw_sc ) ;
            if ( hw_fast ) 
                printf ( " (fast)" ) ;
#if 0
            printf ( "\r\n  hardware (direct) says = 0x" ) ;
            printhex ( (* ( (unsigned *) CONV(i) ))) ;
#endif
            printf ( "\r\n  timer = 0x%08x", TEST_HW_Get_Result_Timer() ) ;
#endif
            printf ( "\r\n  cycle = 0x%08x", cycle ) ;
            printf ( "\r\n" ) ;
            errors ++ ;
            if ( errors > 6 ) break ;
        }
    }
#ifdef __MICROBLAZE__
#if 0 
    for ( i = 0 ; i < MM_SIZE ; i ++ )
    {
        data_t sw_mm = SNOOP ( i ) ;
        data_t hw_mm = (data_t) (* ( (unsigned *) i )) ;

        if ( sw_mm != hw_mm )
        {
            printf ( "Snoop error.\r\n"
                     "  address = 0x" ) ;
            printhex ( i ) ;
            printf ( "\r\n  software says = 0x" ) ;
            printhex ( sw_mm ) ;
            printf ( "\r\n  hardware says = 0x" ) ;
            printhex ( hw_mm ) ;
            printf ( "\r\n" ) ;
            errors ++ ;
            if ( errors > 6 ) break ;
        }
    }
#endif
#endif
    if (( errors != 0 ) && call_exit )
    {
        printf ( "\r\n" ) ;
        printf ( "\r\n" ) ;
        TEST_SW_Dump () ;
        exit ( 1 ) ;
    }
}

static void Status_Exit ( void )
{
    printf ( "\r\nexit - final memory check:\r\n" ) ;
    Check ( 0 ) ;
    printf ( "\r\nopen_ok = 0x" ) ;
    printhex ( open_ok ) ;
    printf ( "\r\nopen_failed = 0x" ) ;
    printhex ( open_failed ) ;
    printf ( "\r\nactive_hwm = 0x" ) ;
    printhex ( active_hwm ) ;
    printf ( "\r\nclose_ok = 0x" ) ;
    printhex ( close_ok ) ;
    printf ( "\r\nunique = 0x" ) ;
    printhex ( unique ) ;
    printf ( "\r\ncycle = 0x" ) ;
    printhex ( cycle ) ;
    printf ( "\r\n\r\nBe seeing you.\r\n" ) ;
}

static int Compare_Free_List_Entries ( const void * p1 , const void * p2 )
{
    const Free_List_Entry * e1 = (const Free_List_Entry *) p1 ;
    const Free_List_Entry * e2 = (const Free_List_Entry *) p2 ;
    if ( e1 -> copy_location < e2 -> copy_location )
    {
        return -1 ;
    } else if ( e1 -> copy_location > e2 -> copy_location )
    {
        return 1 ;
    }
    if ( e1 -> size < e2 -> size )
    {
        return -1 ;
    } else if ( e1 -> size > e2 -> size )
    {
        return 1 ;
    }
    return 0 ;
}

static void Shift_List ( void * target , unsigned size , unsigned number )
{
    unsigned * t2 = (unsigned *) target ;
    unsigned total , j ;

    assert (( size & 3 ) == 0 ) ;
    size /= 4 ;

    total = size * number ;
    for ( j = 0 ; j < total ; j ++ )
    {
        t2 [ j ] = t2 [ j + size ] ;
    }
    for ( j = 0 ; j < size ; j ++ )
    {
        t2 [ j + total ] = 0 ;
    }
}

#ifdef __MICROBLAZE__
void Timing_Test ( void )
{
    unsigned    sum , max , min , i , j , x , r ;
    unsigned    scale = 10 ; /* 20 ;*/
    unsigned    timer , ref_should_be , total = ( 1 << scale ) ;
    char * ops [] =
        { "Slow load" ,     /* 0 */
            "Slow store" ,
            "Open (8)" ,
            "Close (8)" ,
            "Open (16)" ,   /* 4 */
            "Close (16)" ,
            "Open (32)" ,
            "Close (32)" ,
            "Open (64)" ,   /* 8 */
            "Close (64)" ,
            "Open (128)" ,   
            "Close (128)" ,
            "Fast load",    /* 12 */
            "Fast store",   
            "Open_RO (8)" ,
            "Close_RO (8)" ,
            "Open_RO (16)" ,   /* 16 */
            "Close_RO (16)" ,
            "Open_RO (32)" ,
            "Close_RO (32)" ,
            "Open_RO (64)" ,   /* 20 */
            "Close_RO (64)" ,
            "Open_RO (128)" ,   
            "Close_RO (128)" ,
            NULL 
        } ;

    printf ( "test " ) ;
    TEST_HW_LOAD ( 0 ) ;
    printf ( "open " ) ;
    r = TEST_HW_OPEN ( 0 , 8 , 0 ) ;
    printf ( "load " ) ;
    TEST_HW_LOAD ( 0 ) ;
    timer = TEST_HW_Get_Result_Timer () ;
    printf ( "\r\n" ) ;
    TEST_HW_CLOSE ( r ) ;
    printf ( "Fast load time is " ) ;
    printhex ( timer ) ;
    printf ( "\r\n" ) ;
    if ( r != ( NUM_ENTRIES - 1 ) )
    {
        printf ( "Reference is " ) ;
        printhex ( r ) ;
        printf ( "\r\n" ) ;
        printf ( "Reference is wrong!" ) ;
        exit ( 1 ) ;
    }


    for ( i = 0 ; ops[ i ] != NULL ; i ++ )
    {
        printf ( "%s", ops [ i ] ) ;
        max = 0 ;
        min = total ;
        sum = 0 ;

        for ( j = 0 ; ( j < total )
                && ( sum < ( 1 << scale )) ; j ++ )
        {
            ref_should_be = NUM_ENTRIES - 1 ;
            x = rng () % ( NUM_WORDS - 128 ) ; 
            r = 0 ;
            //printf ( "\n+" ) ;
            //printhex ( x ) ;
            //printf ( " " ) ;
            switch ( i )
            {
                case 0 :    //printf ( "?" ) ;
                            TEST_HW_LOAD ( x ) ;
                            //printf ( "!" ) ;
                            break ;
                case 1 :    TEST_HW_STORE ( x , j ) ;
                            break ;
                case 2 :    r = TEST_HW_OPEN ( x , 8 , 0 ) ;
                            break ;
                case 3 :    r = TEST_HW_OPEN ( x , 8 , 0 ) ;
                            TEST_HW_CLOSE ( r ) ;
                            break ;
                case 4 :    r = TEST_HW_OPEN ( x , 16 , 0 ) ;
                            break ;
                case 5 :    r = TEST_HW_OPEN ( x , 16 , 0 ) ;
                            TEST_HW_CLOSE ( r ) ;
                            break ;
                case 6 :    r = TEST_HW_OPEN ( x , 32 , 0 ) ;
                            break ;
                case 7 :    r = TEST_HW_OPEN ( x , 32 , 0 ) ;
                            TEST_HW_CLOSE ( r ) ;
                            break ;
                case 8 :    r = TEST_HW_OPEN ( x , 64 , 0 ) ;
                            break ;
                case 9 :    r = TEST_HW_OPEN ( x , 64 , 0 ) ;
                            TEST_HW_CLOSE ( r ) ;
                            break ;
                case 10 :   r = TEST_HW_OPEN ( x , 128 , 0 ) ;
                            break ;
                case 11 :   r = TEST_HW_OPEN ( x , 128 , 0 ) ;
                            TEST_HW_CLOSE ( r ) ;
                            break ;
                case 12 :   r = TEST_HW_OPEN ( x , 8 , 0 ) ;
                            TEST_HW_LOAD ( x ) ;
                            break ;
                case 13 :   r = TEST_HW_OPEN ( x , 8 , 0 ) ;
                            TEST_HW_STORE ( x , j ) ;
                            break ;
                case 14 :   r = TEST_HW_OPEN_RO ( x , 8 , 0 ) ;
                            ref_should_be = NUM_RO_ENTRIES ;
                            break ;
                case 15 :   r = TEST_HW_OPEN_RO ( x , 8 , 0 ) ;
                            TEST_HW_CLOSE ( r ) ;
                            break ;
                case 16 :   r = TEST_HW_OPEN_RO ( x , 16 , 0 ) ;
                            ref_should_be = NUM_RO_ENTRIES ;
                            break ;
                case 17 :   r = TEST_HW_OPEN_RO ( x , 16 , 0 ) ;
                            TEST_HW_CLOSE ( r ) ;
                            break ;
                case 18 :   r = TEST_HW_OPEN_RO ( x , 32 , 0 ) ;
                            ref_should_be = NUM_RO_ENTRIES ;
                            break ;
                case 19 :   r = TEST_HW_OPEN_RO ( x , 32 , 0 ) ;
                            TEST_HW_CLOSE ( r ) ;
                            break ;
                case 20 :   r = TEST_HW_OPEN_RO ( x , 64 , 0 ) ;
                            ref_should_be = NUM_RO_ENTRIES ;
                            break ;
                case 21 :   r = TEST_HW_OPEN_RO ( x , 64 , 0 ) ;
                            TEST_HW_CLOSE ( r ) ;
                            break ;
                case 22 :   r = TEST_HW_OPEN_RO ( x , 128 , 0 ) ;
                            ref_should_be = NUM_RO_ENTRIES ;
                            break ;
                case 23 :   r = TEST_HW_OPEN_RO ( x , 128 , 0 ) ;
                            TEST_HW_CLOSE ( r ) ;
                            break ;
                default :   assert ( 0 ) ; break ;


            }
            //printf ( "." ) ;
            timer = TEST_HW_Get_Result_Timer () ;
            if ( max < timer ) max = timer ;
            if ( timer < min ) min = timer ;
            //printhex ( timer ) ;
            sum += timer ;
            switch ( i )
            {
                case 2 :    
                case 4 :    
                case 6 :    
                case 8 :    
                case 10 :    
                case 12 :    
                case 13 :   
                case 14 :   
                case 16 :   
                case 18 :   
                case 20 :   
                case 22 :   
                            TEST_HW_CLOSE ( r ) ;
                            if ( r != ref_should_be )
                            {
                                printf ( "Reference is wrong! " ) ;
                                printhex ( r ) ;
                                exit ( 1 ) ;
                            }
                            break ;
                default :   break ;
            }
        }
        printf ( " sum " ) ;
        printhex ( sum ) ;
        printf ( " total " ) ;
        printhex ( j ) ;
        printf ( " min " ) ;
        printhex ( min ) ;
        printf ( " max " ) ;
        printhex ( max ) ;
        printf ( "\r\n" ) ;
    }
}
#endif

static inline int Fast_Rand ( void )
{
    return rng () ;
}

static void Main_Test ( unsigned max_cycles )
{
    unsigned num_ro_open, num_rw_open;
    unsigned i, limit ;

    num_ro_open = num_rw_open = 0;

    /* Clear memory */
    memset ( active_list , 0 , ( NUM_RO_ENTRIES + NUM_RW_ENTRIES ) *
                            sizeof ( Free_List_Entry ) ) ;
    memset ( free_list , 0 , MAX_FREE_LIST_SIZE *
                            sizeof ( Free_List_Entry ) ) ;
    memset ( SHADOW , 0 , sizeof ( data_t ) * MM_SIZE ) ;
    for ( i = 0 ; i < MM_SIZE ; i ++ )
    {
#ifdef __MICROBLAZE__
        TEST_HW_STORE ( i , 0 ) ;
#endif
        TEST_SW_STORE ( i , 0 ) ;
    }

    /* Initial test */
    Check ( 1 ) ;

    /* Initial free list */
    free_list_size = 1 ;
    free_list [ 0 ] . size = SP_SIZE ;
    free_list [ 0 ] . copy_location = 0 ;
    active_list_size = 0;

    /* Go */
    limit = max_cycles + cycle;
    for ( ; cycle < limit ; cycle ++ )
    {
        unsigned r = Fast_Rand () ;

        if (( active_list_size < ( NUM_RW_ENTRIES + NUM_RO_ENTRIES ))
        && (( r & 3 ) != 0 ))
        {
            /* Open (Allocate space first) */
            int     read_only ;
            int     attempts = 0;
            size_t  size = (( Fast_Rand() % (( SP_SIZE / 4 ) - 4 )) 
                                    >> (( r >> 4 ) & 3 )) ;
            int     ok = 0;

            if ( num_ro_open >= NUM_RO_ENTRIES )
            {
                read_only = 0 ;
            } else if ( num_rw_open >= NUM_RW_ENTRIES )
            {
                read_only = 1 ;
            } else {
                read_only = ! ( r & 0x8 ) ;
            }
            do {
                size_t  address ;
                
                size ++;
                address = ( Fast_Rand() % ( MM_SIZE - size )) ;
                for ( i = 0 ; i < free_list_size ; i ++ )
                {
                    if ( free_list [ i ] . size < size )
                    {
                        continue ;
                    }
                    /* Found a free space */
                    size_t copy_location = free_list [ i ] . copy_location ;
                    static_cache_ref ref ; 
                    static_cache_ref hw_ref ;

                    if ( read_only )
                    {
                        assert ( num_ro_open < NUM_RO_ENTRIES ) ;
                        num_ro_open ++ ;
                        ref = TEST_SW_OPEN_RO ( address , size , copy_location ) ;
                        assert ( ref > 0 ) ;
                        assert ( ref < FIRST_RW_ENTRY ) ;
                    } else {
                        assert ( num_rw_open < NUM_RW_ENTRIES ) ;
                        num_rw_open ++ ;
                        ref = TEST_SW_OPEN ( address , size , copy_location ) ;
                        assert ( ref >= FIRST_RW_ENTRY ) ;
                        assert ( ref < NUM_ENTRIES ) ;
                    }
                    hw_ref = ref ; 
#ifdef __MICROBLAZE__
#ifdef VERBOSE
                    unsigned c = HW_Get_Result_Counter () ;
#endif
                    if ( read_only )
                    {
                        hw_ref = TEST_HW_OPEN_RO ( address , 
                                    size , copy_location ) ;
                    } else {
                        hw_ref = TEST_HW_OPEN ( address , 
                                    size , copy_location ) ;
                    }
#ifdef VERBOSE
                    printhex ( cycle ) ;
                    printf ( ": OPEN " ) ;
                    printhex ( address ) ;
                    printf ( " sz " ) ;
                    printhex ( size << 0 ) ;
                    printf ( " cl " ) ;
                    printhex ( copy_location ) ;
                    printf ( " rc " ) ;
                    printhex ( HW_Get_Result_Counter () - c ) ;
                    printf ( " ref " ) ;
                    printhex ( hw_ref ) ;
                    printf ( " timer " ) ;
                    printhex ( TEST_HW_Get_Result_Timer () ) ;
                    printf ( "\r\n" ) ;
#endif
#endif

                    if ( ref != hw_ref )
                    {
                        printhex ( cycle ) ;
                        printf ( " software reference is " ) ;
                        printhex ( ref ) ;
                        printf ( " and hardware reference is " ) ;
                        printhex ( hw_ref ) ;
                        printf ( "\r\n" ) ;
                        exit ( 1 );
                    }
                    active_list [ active_list_size ] . ref = ref ;
                    active_list [ active_list_size ] . size = size ;
                    active_list [ active_list_size ] . 
                                    copy_location = copy_location ;
                    active_list [ active_list_size ] . address = address ;
                    active_list_size ++ ;
                    free_list [ i ] . size -= size ;
                    free_list [ i ] . copy_location += size ;
                    ok = 1 ;
                    break ;
                }
                if ( ! ok )
                {
                    size /= 2;
                    attempts ++;
                }
            } while (( ! ok ) && ( attempts < 10 ));

            if ( ok )
            {
                open_ok ++ ;
                if ( active_list_size > active_hwm )
                {
                    active_hwm = active_list_size ;
                }
            } else {
                open_failed ++ ;
            }
        } else if ( active_list_size > 0 )
        {
            /* Close (and deallocate space) */
            unsigned            victim = Fast_Rand () % active_list_size ;
            int                 notify = 0 ; /* debugging control, e.g. ( cycle > 0x2263a ) ; */
            static_cache_ref    ref = active_list [ victim ] . ref ;

           
            if ( ref < FIRST_RW_ENTRY )
            {
                assert ( ref != 0 ) ;
                assert ( num_ro_open != 0 ) ;
                num_ro_open -- ;
            } else {
                assert ( ref < NUM_ENTRIES ) ;
                assert ( num_rw_open != 0 ) ;
                num_rw_open -- ;
            }
            TEST_SW_CLOSE ( ref ) ;
#ifdef __MICROBLAZE__
#ifdef VERBOSE
            printhex ( cycle ) ;
            printf ( ": CLOSE " ) ;
            printf ( " ref " ) ;
            printhex ( ref ) ;
            printf ( "\r\n" ) ;
#endif
            TEST_HW_CLOSE ( ref ) ;
#endif

            /* Deallocation: add to free list */
            assert (( free_list_size + 1 ) < MAX_FREE_LIST_SIZE ) ;
            free_list [ free_list_size ] . size = 
                        active_list [ victim ] . size ;
            free_list [ free_list_size ] . copy_location = 
                        active_list [ victim ] . copy_location ;
            free_list_size ++ ;

            /* Removal from active list */
            if (( victim + 1 ) < active_list_size )
            {
                Shift_List ( & active_list [ victim ] ,
                            sizeof ( Active_List_Entry ) ,
                            (( active_list_size - victim ) - 1 ) ) ;
            }
            active_list_size -- ;
            close_ok ++ ;

            /* Merge adjacent free regions in free list */
            qsort ( free_list , free_list_size ,
                            sizeof ( Free_List_Entry ) ,
                            Compare_Free_List_Entries ) ;

            for ( i = 0 ; i < ( free_list_size - 1 ) ; i ++ )
            {
                assert ( free_list [ i ] . copy_location <=
                        free_list [ i + 1 ] . copy_location ) ;
                if ( notify )
                {
                    printhex ( i ) ;
                    printf ( " cl: " ) ;
                    printhex ( free_list [ i ] . copy_location ) ;
                    printf ( " sz: " ) ;
                    printhex ( free_list [ i ] . size ) ;
                    printf ( "\r\n" ) ;
                }
            }

            for ( i = 0 ; i < ( free_list_size - 1 ) ; )
            {
                if ( notify )
                {
                    printhex ( i ) ;
                    printf ( "\r\n" ) ;
                }
                assert (( free_list [ i ] . size + 
                        free_list [ i ] . copy_location ) <=
                        free_list [ i + 1 ] . copy_location ) ;
                if (( free_list [ i ] . size + 
                        free_list [ i ] . copy_location ) ==
                        free_list [ i + 1 ] . copy_location )
                {
                    /* Adjacent */
                    if ( notify )
                    {
                        printf ( "shuffle\r\n" ) ;
                    }

                    free_list [ i ] . size += 
                         free_list [ i + 1 ] . size ;
                    free_list [ i + 1 ] . size = ~0 ;
                    free_list [ i + 1 ] . copy_location = ~0 ;
                    if (( i + 2 ) < free_list_size )
                    {
                        if ( notify )
                        {
                            printf ( "shift\r\n" ) ;
                        }
                        Shift_List ( & free_list [ i + 1 ] ,
                            sizeof ( Free_List_Entry ) ,
                            ( free_list_size - ( i + 2 ))) ;
                    }
                    free_list_size -- ;
                } else {
                    i ++ ;
                }
            }
            //Check ( 1 ) ;
        }
        unique += ( r >> 8 ) % 256 ;
        if ((( cycle % 64 ) == 0 )
        && (( unique % 8 ) == 0 ))
        {
            /* Cache flush operation */
#if XPAR_MICROBLAZE_0_USE_ICACHE
            microblaze_disable_icache();
            microblaze_enable_icache();
#endif
        }
        for ( i = 0 ; i < 10 ; i ++ )
        {
            size_t address = Fast_Rand () % MM_SIZE ; 
            int hw_fast, sw_fast ;
            unsigned timer = 0 ;

            TEST_SW_STORE ( address , unique ) ;
            hw_fast = sw_fast = 
                    TEST_SW_Get_Result_Timer () < TIMER_THRESHOLD ;
#ifdef __MICROBLAZE__
            TEST_HW_STORE ( address , unique ) ;
            timer = TEST_HW_Get_Result_Timer () ;
            hw_fast = timer < TIMER_THRESHOLD ;
#endif
            SHADOW [ address ] = unique ;
#ifdef VERBOSE
            printhex ( address ) ;
            printf ( " := " ) ;
            printhex ( unique ) ;
            printf ( "\r\n" ) ;
#endif
            if (( !! hw_fast ) != ( !! sw_fast ))
            {
                printf ( "STORE timing inconsistency at address " ) ;
                printhex ( address ) ;
                printf ( " hw timer " ) ;
                printhex ( timer ) ;
                printf ( " sw says " ) ;
                if ( sw_fast )
                {
                    printf ( "fast\r\n" ) ;
                } else {
                    printf ( "slow\r\n" ) ;
                }
                exit ( 1 ) ;
            }
            unique ++ ;
            unique += i ; 
        }
        if (( cycle % 64 ) == 0 )
        {
            Check ( 1 ) ;
        }
        if (( cycle % 256 ) == 255 )
        {
            //unsigned x ;
            //asm volatile ( "mfs %0, rPC\n" : "=r"(x) ) ;
            printf ( "cycle " ) ;
            printhex ( cycle ) ;
            printf ( " " ) ;
            printhex ( unique ) ;
            printf ( " " ) ;
            if (( cycle % 2048 ) == 255 )
            {
                printhex ( Fast_Rand () ) ;
                printf ( "\r\n" ) ;
                TEST_SW_Dump () ;
            } else {
                printf ( "\r\n" ) ;
            }
        }
    }
    /* Clear out table */
    for ( i = 0 ; i < active_list_size ; i ++ )
    {
        static_cache_ref ref = active_list [ i ] . ref ;
        TEST_SW_CLOSE ( ref ) ;
#ifdef __MICROBLAZE__
        TEST_HW_CLOSE ( ref ) ;
#endif
    }
    Check ( 1 ) ;
}

int Towering ( void ) ;

void Test ( void )
{

    /* Memory */
    assert (NUM_ENTRIES != 0);
    memset ( SHADOW , 0 , sizeof ( data_t ) * MM_SIZE ) ;
    free_list = calloc ( MAX_FREE_LIST_SIZE , sizeof ( Free_List_Entry ) ) ;
    active_list = calloc ( NUM_RO_ENTRIES + NUM_RW_ENTRIES , 
                        sizeof ( Free_List_Entry ) ) ;
    assert (free_list != NULL);
    assert (active_list != NULL);
    memset ( SHADOW , 0 , sizeof ( data_t ) * MM_SIZE ) ;
    unique = 0 ;
    cycle = 0;

    Cache_Test () ;
    reset_rng () ;
    printf ( "Timing test:\r\n" ) ;

#ifdef __MICROBLAZE__
    Timing_Test () ;
#endif

    atexit ( Status_Exit ) ;

    reset_rng () ;
    printf ( "Mini test 1:\r\n" ) ;
    Main_Test ( 1000 ) ;
    printf ( "Mini test 2:\r\n" ) ;
    Main_Test ( 1000 ) ;

    printf ( "'Towering' test:\r\n" ) ;
    reset_rng () ;
    if ( ! Towering () )
    {
        return ;
    }

    printf ( "Full test:\r\n" ) ;
    Main_Test ( NUM_CYCLES ) ;

    printf ( "All the tests have passed!\r\n" ) ;
}

#ifdef __MICROBLAZE__
#else
int main ( void )
{
    Test () ;
}

#endif

