/*
 * Scratchpad MMU - Test program
 * Copyright (C) Jack Whitham 2010
 * http://www.jwhitham.org.uk/c/smmu.html
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation
 * (version 2.1 of the License only).
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */


#include <string.h>
#include <assert.h>

#include "hwdriver.h"
#include "simif.h"
#include "smmu.h"
#include "mb_core.h"

#define HW_SCRAMBLE_KEY     0x2354f98aUL
#define SW_SCRAMBLE_KEY     0xf73df2d9UL

static unsigned hw_sram [ NUM_WORDS ] ;
static unsigned sw_sram [ NUM_WORDS ] ;
static unsigned sw_last_result;
static unsigned hw_last_result;
static struct SMMU_Context_struct * smmu_sys ;
static MB_Context * mbc ;
static struct MB_System_Context_struct * sys ;

static unsigned MB_Load ( void * m_user , unsigned address , unsigned size )
{
    mbc->clock_cycle += 100;
    address &= 0xffff;
    address /= 4;
    return sw_sram[ address ] ^ SW_SCRAMBLE_KEY;
}

static void MB_Store ( void * m_user , unsigned address , 
        unsigned data , unsigned size )
{
    assert(size == 4);
    mbc->clock_cycle += 100;
    address &= 0xffff;
    address /= 4;
    sw_sram[ address ] = SW_SCRAMBLE_KEY ^ data;
}

static void MB_Trace ( void * t_user , struct MB_Context_struct * mc ,
        MB_Trace_Name trace_name , const void * param )
{
}

void Simulator_Init ( void )
{
    unsigned i;

    for (i = 0; i < NUM_WORDS; i++)
    {
        sw_sram[ i ] = SW_SCRAMBLE_KEY;
        hw_sram[ i ] = HW_SCRAMBLE_KEY ^ i;
    }
    sw_last_result = 0;
    hw_last_result = 0;
    sys = MB_System_Init ( MB_Trace , 100 , NULL ) ;
    mbc = MB_System_Get_MB_Context ( sys ) ;

    mbc->load_fn = MB_Load;
    mbc->store_fn = MB_Store;
    mbc->trace_fn = MB_Trace;

    assert(NUM_ENTRIES != 0);

    smmu_sys = SMMU_Init ( sys , 
            NUM_RW_ENTRIES , NUM_RO_ENTRIES , SP_SIZE ) ;
    assert(smmu_sys != NULL);
}

unsigned TEST_SW_OPEN(unsigned address, unsigned size, unsigned copy_loc)
{
    return SMMU_Open(smmu_sys, (address * 4) | ACTIVE_AREA, 
                    size, copy_loc, 0, 0, 0);
}

unsigned TEST_SW_OPEN_RO(unsigned address, unsigned size, unsigned copy_loc)
{
    return SMMU_Open(smmu_sys, (address * 4) | ACTIVE_AREA, 
                    size, copy_loc, 0, 0, 1);
}

unsigned TEST_HW_OPEN(unsigned address, unsigned size, unsigned copy_loc)
{
    unsigned ref;
    assert((address + size) <= NUM_WORDS);
    assert((copy_loc + size) <= SP_SIZE);
    ref = HW_OPEN(&hw_sram[ address ], size << 2, copy_loc);
    hw_last_result = HW_Get_Result_Timer();

    return ref;
}

unsigned TEST_HW_OPEN_RO(unsigned address, unsigned size, unsigned copy_loc)
{
    unsigned ref;
    assert((address + size) <= NUM_WORDS);
    assert((copy_loc + size) <= SP_SIZE);
    ref = HW_OPEN_RO(&hw_sram[ address ], size << 2, copy_loc);
    hw_last_result = HW_Get_Result_Timer();
    return ref;
}

unsigned TEST_SW_LOAD(unsigned address)
{
    unsigned data, start;
    
    start = mbc->clock_cycle;
    data = mbc->load_fn(smmu_sys, (address * 4) | ACTIVE_AREA, 4);
    sw_last_result = mbc->clock_cycle - start;
    return data;
}

void TEST_SW_STORE(unsigned address, unsigned data)
{
    unsigned start;
    
    start = mbc->clock_cycle;
    mbc->store_fn(smmu_sys, (address * 4) | ACTIVE_AREA, data, 4);
    sw_last_result = mbc->clock_cycle - start;
}

unsigned TEST_HW_LOAD(unsigned address)
{
    unsigned data;
    assert(address < NUM_WORDS);
    asm volatile (
            "lwi    %1, %2, 0\n" 
            "lwi    %0, %3, 0\n"
            : "=&r"(hw_last_result), "=&r"(data)
            : "r"((& hw_sram[ address ] )), "r"((CONTROL)) ) ;
    return data ^ HW_SCRAMBLE_KEY ^ address;
}

void TEST_HW_STORE(unsigned address, unsigned data)
{
    assert(address < NUM_WORDS);
    data = data ^ HW_SCRAMBLE_KEY ^ address ;

    asm volatile (
            "swi    %1, %2, 0\n" 
            "lwi    %0, %3, 0\n"
            : "=&r"(hw_last_result)
            : "r"(data), 
                "r"((& hw_sram[ address ] )), "r"((CONTROL)) ) ;
}

void TEST_SW_CLOSE(unsigned ref)
{
    SMMU_Close(smmu_sys, ref);
}

void TEST_HW_CLOSE(unsigned ref)
{
    HW_CLOSE(ref);
    hw_last_result = HW_Get_Result_Timer();
}

unsigned TEST_SW_Get_Result_Timer(void)
{
    return sw_last_result;
}

unsigned TEST_HW_Get_Result_Timer(void)
{
    return hw_last_result & 0xffff;
}

void TEST_SW_Dump(void)
{
    SMMU_Dump_Table(smmu_sys);
}



